#############################################################################
#
# Copyright (c) 2006 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""catalog package test runner

"""
import doctest
import re
import unittest

import BTrees.Interfaces
import BTrees.LFBTree
import BTrees.LOBTree
import BTrees.OLBTree
import zope.component.factory
import zope.component.interfaces
import zope.component.testing
from zope.testing import module
from zope.testing import renormalizing

import zc.catalog
import zc.catalog.interfaces
from zc.catalog import catalogindex
from zc.catalog import extentcatalog
from zc.catalog import globber
from zc.catalog import index
from zc.catalog import stemmer


class TestAbstractIndex(unittest.TestCase):

    def test_family_on_cls(self):
        self.assertIsInstance(index.AbstractIndex.family,
                              index.FamilyProperty)

    def test_clear_cruft(self):
        i = index.AbstractIndex()
        i.__dict__['BTreeAPI'] = None
        del i.__dict__['family']
        self.assertIn('BTreeAPI', i.__dict__)
        getattr(i, 'family')
        self.assertNotIn('BTreeAPI', i.__dict__)

    def test_family(self):
        class Family:
            class OO:
                class BTree:
                    pass
            IO = OO
        i = index.AbstractIndex(family=Family)
        self.assertIs(i.family, Family)

    def test_empty_values(self):
        i = index.AbstractIndex()
        res = i.values(doc_id=1)
        self.assertEqual((), res)


class TestValueIndex(unittest.TestCase):

    def test_empty_values(self):
        i = index.ValueIndex()
        res = i.values(doc_id=1)
        self.assertEqual((), res)


class TestSetIndex(unittest.TestCase):

    def test_removed(self):
        i = index.SetIndex()
        i.index_doc(1, ('foo', 'bar'))
        i.index_doc(1, ('foo',))

        self.assertEqual(1, i.wordCount.value)

    def test_appy_all_of_empty(self):
        i = index.SetIndex()
        res = i.apply({'all_of': ()})
        self.assertEqual(len(res), 0)


class TestNormalizationWrapper(unittest.TestCase):

    def test_pass_to_index(self):
        i = index.SetIndex()

        class Normaziler:
            @classmethod
            def value(cls, v):
                return v
        n = index.NormalizationWrapper(i, Normaziler)

        self.assertEqual(i.documentCount(), n.documentCount())
        self.assertEqual(i.wordCount(), n.wordCount())
        n.clear()

        n.index_doc(1, ('foo',))
        self.assertEqual(i.wordCount(), n.wordCount())

        self.assertEqual(n.containsValue('foo'), i.containsValue('foo'))


class TestExtent(unittest.TestCase):

    def test_BTreeAPI(self):
        i = extentcatalog.Extent()
        self.assertIsNotNone(i.BTreeAPI)

    def test_bool(self):
        i = extentcatalog.Extent()
        self.assertFalse(i)
        i.add(1, None)
        self.assertTrue(i)
        self.assertEqual(1, len(i))

    def test_discard_missing(self):
        i = extentcatalog.Extent()
        i.discard(0)
        self.assertEqual(0, len(i))

    def test_catalog_update(self):
        from zope.interface.interfaces import ComponentLookupError
        c = extentcatalog.Catalog(extentcatalog.Extent())
        i = index.SetIndex()
        i.__parent__ = None
        self.assertRaises(ComponentLookupError, c.updateIndex, i)


class TestGlob(unittest.TestCase):

    def test_bad_parse(self):
        class Lexicon:
            pass
        res = globber.glob('', Lexicon())
        self.assertIsNone(res)


class TestCatalogIndex(unittest.TestCase):

    def test_datetimevalueindex(self):
        i = catalogindex.DateTimeValueIndex(field_name='foo')
        self.assertTrue(zc.catalog.interfaces.IValueIndex.providedBy(i))

    def test_datetimesetindex(self):
        i = catalogindex.DateTimeSetIndex(field_name='foo')
        self.assertTrue(zc.catalog.interfaces.ISetIndex.providedBy(i))


@unittest.skipUnless(stemmer.broken, "Only for broken stemmers")
class TestBrokenStemmer(unittest.TestCase):

    def test_broken(self):
        s = stemmer.Stemmer()
        self.assertIs(stemmer.broken, s.stemmer)
        self.assertEqual('word', s.stemmer.stem("word"))


def setUp32bit(test):
    zope.component.testing.setUp(test)
    test.globs["btrees_family"] = BTrees.family32


def modSetUp32bit(test):
    setUp32bit(test)
    module.setUp(test, 'zc.catalog.doctest_test')


def setUp64bit(test):
    zope.component.testing.setUp(test)
    test.globs["btrees_family"] = BTrees.family64


def modSetUp64bit(test):
    setUp64bit(test)
    module.setUp(test, 'zc.catalog.doctest_test')


def tearDown(test):
    zope.component.testing.tearDown(test)


def modTearDown(test):
    module.tearDown(test)
    zope.component.testing.tearDown(test)


def test_suite():
    checker = renormalizing.RENormalizing((
        (re.compile(r"<class 'BTrees."), "<type 'BTrees."),
        (re.compile(r"<module 'BTrees\._"), "<module 'BTrees."),
    ))
    tests = unittest.TestSuite((
        # 32 bits
        doctest.DocFileSuite(
            'extentcatalog.rst', setUp=modSetUp32bit, tearDown=modTearDown),
        doctest.DocFileSuite(
            'setindex.rst', setUp=setUp32bit, tearDown=tearDown),
        doctest.DocFileSuite(
            'valueindex.rst', setUp=setUp32bit, tearDown=tearDown),
        doctest.DocFileSuite(
            'normalizedindex.rst', setUp=setUp32bit, tearDown=tearDown),
        doctest.DocFileSuite(
            'globber.rst', setUp=setUp32bit, tearDown=tearDown),
        doctest.DocFileSuite(
            'callablewrapper.rst', setUp=setUp32bit, tearDown=tearDown),

        # 64 bits
        doctest.DocFileSuite(
            'extentcatalog.rst', setUp=modSetUp64bit, tearDown=modTearDown),
        doctest.DocFileSuite('setindex.rst', setUp=setUp64bit,
                             tearDown=tearDown),
        doctest.DocFileSuite('valueindex.rst', setUp=setUp64bit,
                             tearDown=tearDown),
        doctest.DocFileSuite('normalizedindex.rst', setUp=setUp64bit,
                             tearDown=tearDown),
        doctest.DocFileSuite('globber.rst', setUp=setUp64bit,
                             tearDown=tearDown),
        doctest.DocFileSuite('callablewrapper.rst', setUp=setUp64bit,
                             tearDown=tearDown),

        # legacy data support
        doctest.DocFileSuite(
            'legacy.rst',
            optionflags=doctest.ELLIPSIS,
            checker=checker),
    ))

    if not stemmer.broken:  # pragma: no cover
        tests.addTest(doctest.DocFileSuite('stemmer.rst'))

    tests.addTest(unittest.defaultTestLoader.loadTestsFromName(__name__))
    return tests



##############################################################################
#
# Copyright (c) 2004 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Functional tests for `zc.catalog.browser`.

"""
import doctest

import transaction
import zope.app.appsetup.bootstrap
import zope.intid
import zope.intid.interfaces
from zope.app.wsgi.testlayer import BrowserLayer
from zope.processlifetime import IDatabaseOpenedWithRoot
from zope.testbrowser.wsgi import TestBrowserLayer

import zc.catalog.browser


class _ZcCatalogLayer(TestBrowserLayer,
                      BrowserLayer):
    pass


ZcCatalogLayer = _ZcCatalogLayer(zc.catalog.browser)


@zope.component.adapter(IDatabaseOpenedWithRoot)
def initializeIntIds(event):
    _db, connection, _root, root_folder = (
        zope.app.appsetup.bootstrap.getInformationFromEvent(event))
    sm = root_folder.getSiteManager()
    if 'test-int-ids' not in sm['default']:
        intids = zope.intid.IntIds()

        sm["default"]["test-int-ids"] = intids
        sm.registerUtility(
            intids,
            zope.intid.interfaces.IIntIds)
        transaction.commit()
    connection.close()


def test_suite():
    suite = doctest.DocFileSuite(
        "README.rst",
        optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)
    suite.layer = ZcCatalogLayer
    return suite


class LoginLogout:
    # dummy to avoid dep on zope.app.security
    def __call__(self):
        return



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Setup for zc.catalog package

"""
import os

from setuptools import find_packages
from setuptools import setup


def read(*rnames):
    with open(os.path.join(os.path.dirname('.'), *rnames)) as f:
        return f.read()


stemmer_requires = [
    'zopyx.txng3.ext >= 2.0.0',
]

tests_require = [
    'ZODB',
    'beautifulsoup4',
    'zope.annotation',
    'zope.app.appsetup',
    'zope.app.basicskin',
    'zope.app.catalog',
    'zope.app.catalog',
    'zope.app.component',
    'zope.app.container',
    'zope.app.form',
    'zope.app.publisher >= 4.0',
    'zope.app.rotterdam',
    'zope.app.schema >= 4.0',
    'zope.app.wsgi',
    'zope.browsermenu',
    'zope.browserpage',
    'zope.browserresource',
    'zope.dottedname',
    'zope.keyreference',
    'zope.login',
    'zope.password',
    'zope.principalannotation',
    'zope.principalregistry',
    'zope.securitypolicy',
    'zope.testbrowser >= 5.2',
    'zope.testing',
    'zope.testrunner',
] + stemmer_requires

setup(name='zc.catalog',
      version='4.0',
      author='Zope Corporation and Contributors',
      author_email='zope-dev@zope.org',
      description="Extensions to the Zope 3 Catalog",
      long_description=(
          read('README.rst')
          + '\n\n.. contents::\n\n' +
          read('CHANGES.rst')
          + '\n\n' +
          read('src', 'zc', 'catalog', 'valueindex.rst')
          + '\n\n' +
          read('src', 'zc', 'catalog', 'setindex.rst')
          + '\n\n' +
          read('src', 'zc', 'catalog', 'normalizedindex.rst')
          + '\n\n' +
          read('src', 'zc', 'catalog', 'extentcatalog.rst')
          + '\n\n' +
          read('src', 'zc', 'catalog', 'stemmer.rst')
          + '\n\n' +
          read('src', 'zc', 'catalog', 'legacy.rst')
          + '\n\n' +
          read('src', 'zc', 'catalog', 'globber.rst')
          + '\n\n' +
          read('src', 'zc', 'catalog', 'callablewrapper.rst')
          + '\n\n' +
          read('src', 'zc', 'catalog', 'browser', 'README.rst')
      ),
      keywords="zope3 i18n date time duration catalog index",
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Web Environment',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: Zope Public License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: 3.11',
          'Programming Language :: Python :: 3.12',
          'Programming Language :: Python :: 3.13',
          'Programming Language :: Python :: Implementation :: CPython',
          'Programming Language :: Python :: Implementation :: PyPy',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Topic :: Internet :: WWW/HTTP',
          'Framework :: Zope :: 3',
      ],
      url='http://github.com/zopefoundation/zc.catalog',
      license='ZPL 2.1',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      namespace_packages=['zc'],
      python_requires='>=3.9',
      extras_require={
          'test': tests_require,
          'browser': [
              'zope.app.form',
              'zope.browsermenu',
          ],
          'stemmer': stemmer_requires,
      },
      tests_require=tests_require,
      install_requires=[
          'BTrees >= 4.4.1',
          'persistent',
          'pytz',
          'setuptools',
          'zope.catalog >= 4.2.1',
          'zope.component >= 4.3.0',
          'zope.container >= 4.1.0',
          'zope.i18nmessageid >= 4.1.0',
          'zope.index >= 4.3.0',
          'zope.interface >= 4.4.0',
          'zope.intid >= 4.2.0',
          'zope.publisher >= 3.12',
          'zope.schema >= 4.4.2',
          'zope.security >= 4.1.0',
      ],
      include_package_data=True,
      zip_safe=False)



#############################################################################
#
# Copyright (c) 2006 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""extent catalog

"""

import sys

import BTrees
import persistent
import zope.component
from zope import interface
from zope.catalog import catalog
from zope.intid.interfaces import IIntIds

from zc.catalog import interfaces


@interface.implementer(interfaces.IExtent)
class Extent(persistent.Persistent):

    __parent__ = None
    family = BTrees.family32

    def __init__(self, family=None):
        if family is not None:
            self.family = family
        self.set = self.family.IF.TreeSet()

    # Deprecated.
    @property
    def BTreeAPI(self):
        return sys.modules[self.set.__class__.__module__]

    def __len__(self):
        return len(self.set)

    def add(self, uid, obj):
        self.set.insert(uid)

    def clear(self):
        self.set.clear()

    def __or__(self, other):
        "extent | set"
        return self.union(other)

    __ror__ = __or__

    def union(self, other, self_weight=1, other_weight=1):
        return self.family.IF.weightedUnion(
            self.set, other, self_weight, other_weight)[1]

    def __and__(self, other):
        "extent & set"
        return self.intersection(other)

    __rand__ = __and__

    def intersection(self, other, self_weight=1, other_weight=1):
        return self.family.IF.weightedIntersection(
            self.set, other, self_weight, other_weight)[1]

    def __sub__(self, other):
        "extent - set"
        return self.difference(other)

    def difference(self, other):
        return self.family.IF.difference(self.set, other)

    def __rsub__(self, other):
        "set - extent"
        return self.rdifference(other)

    def rdifference(self, other):
        return self.family.IF.difference(other, self.set)

    def __iter__(self):
        return iter(self.set)

    def __nonzero__(self):
        return bool(self.set)

    __bool__ = __nonzero__

    def __contains__(self, uid):
        return uid in self.set

    def remove(self, uid):
        self.set.remove(uid)

    def discard(self, uid):
        try:
            self.set.remove(uid)
        except KeyError:
            pass


@interface.implementer(interfaces.IFilterExtent)
class FilterExtent(Extent):

    def __init__(self, filter, family=None):
        super().__init__(family=family)
        self.filter = filter

    def add(self, uid, obj):
        if not self.addable(uid, obj):
            raise ValueError
        else:
            self.set.insert(uid)

    def addable(self, uid, obj):
        return self.filter(self, uid, obj)


@interface.implementer(interfaces.ISelfPopulatingExtent)
class NonPopulatingExtent(Extent):
    """Base class for populating extent.

    This simple, no-op implementation comes in handy surprisingly often
    for catalogs that handle a very contained domain within an application.
    """

    populated = False

    def populate(self):
        self.populated = True


@interface.implementer(interfaces.IExtentCatalog)
class Catalog(catalog.Catalog):

    UIDSource = None

    def __init__(self, extent, UIDSource=None):
        """Construct a catalog based on an extent.

        Note that the `family` keyword parameter of the base class
        constructor is not supported here; the family of the extent is
        used.

        """

        self.UIDSource = UIDSource

        if extent.__parent__ is not None:  # pragma: no cover
            raise ValueError("extent's __parent__ must be None")
        super().__init__(family=extent.family)
        self.extent = extent
        extent.__parent__ = self  # inform extent of catalog

    def _getUIDSource(self):
        res = self.UIDSource
        if res is None:
            res = zope.component.getUtility(IIntIds)
        return res

    def clear(self):
        self.extent.clear()
        super().clear()

    def index_doc(self, docid, texts):
        """Register the data in indexes of this catalog.
        """
        try:
            self.extent.add(docid, texts)
        except ValueError:
            self.unindex_doc(docid)
        else:
            super().index_doc(docid, texts)

    def unindex_doc(self, docid):
        if docid in self.extent:
            super().unindex_doc(docid)
            self.extent.remove(docid)

    def searchResults(self, **kwargs):
        res = super().searchResults(**kwargs)
        if res is not None:
            res.uidutil = self._getUIDSource()
        return res

    def updateIndex(self, index):
        if index.__parent__ is not self:
            # not an index in us.  Let the superclass handle it.
            super().updateIndex(index)
        else:
            uidutil = self._getUIDSource()

            if interfaces.ISelfPopulatingExtent.providedBy(self.extent):
                if not self.extent.populated:
                    self.extent.populate()
                    assert self.extent.populated

                for uid in self.extent:
                    obj = uidutil.getObject(uid)
                    index.index_doc(uid, obj)

            else:
                for uid in uidutil:
                    obj = uidutil.getObject(uid)
                    try:
                        self.extent.add(uid, obj)
                    except ValueError:
                        self.unindex_doc(uid)
                    else:
                        index.index_doc(uid, obj)

    def updateIndexes(self):
        uidutil = self._getUIDSource()

        if interfaces.ISelfPopulatingExtent.providedBy(self.extent):
            if not self.extent.populated:
                self.extent.populate()
                assert self.extent.populated

            for uid in self.extent:
                self.index_doc(uid, uidutil.getObject(uid))

        else:
            for uid in uidutil:
                self.index_doc(uid, uidutil.getObject(uid))
